/*
 * Copyright 2017 The OpenSSL Project Authors. All Rights Reserved.
 * Copyright 2017 Ribose Inc. All Rights Reserved.
 *
 * Licensed under the OpenSSL license (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://www.openssl.org/source/license.html
 */

#include "internal/cryptlib.h"

#ifndef OPENSSL_NO_SM3
# include <openssl/evp.h>
# include "crypto/evp.h"
# include "openssl/sm3.h"

static int init(EVP_MD_CTX *ctx)
{
    return sm3_init(EVP_MD_CTX_md_data(ctx));
}

static int update(EVP_MD_CTX *ctx, const void *data, size_t count)
{
    return sm3_update(EVP_MD_CTX_md_data(ctx), data, count);
}

static int final(EVP_MD_CTX *ctx, unsigned char *md)
{
    return sm3_final(md, EVP_MD_CTX_md_data(ctx));
}

static int ctrl(EVP_MD_CTX *ctx, int cmd, int mslen, void *ms)
{
    unsigned char padtmp[32];	/* TODO: Is 32 right? */
    unsigned char sm3tmp[SM3_DIGEST_LENGTH];

    SM3_CTX *sm3;

    if (cmd != EVP_CTRL_SSL3_MASTER_SECRET)
        return -2;

    if (ctx == NULL)
        return 0;

    sm3 = EVP_MD_CTX_md_data(ctx);

    /* SSLv3 client auth handling: see RFC-6101 5.6.8 */
    if (mslen != 48)
        return 0;

    /* At this point hash contains all handshake messages, update
     * with master secret and pad_1.
     */

    if (sm3_update(sm3, ms, mslen) <= 0)
        return 0;

    /* Set padtmp to pad_1 value */
    memset(padtmp, 0x36, sizeof(padtmp));

    if (!sm3_update(sm3, padtmp, sizeof(padtmp)))
        return 0;

    if (!sm3_final(sm3tmp, sm3))
        return 0;

    /* Reinitialise context */

    if (!sm3_init(sm3))
        return 0;

    if (sm3_update(sm3, ms, mslen) <= 0)
        return 0;

    /* Set padtmp to pad_2 value */
    memset(padtmp, 0x5c, sizeof(padtmp));

    if (!sm3_update(sm3, padtmp, sizeof(padtmp)))
        return 0;

    if (!sm3_update(sm3, sm3tmp, sizeof(sm3tmp)))
        return 0;

    /* Now when ctx is finalised it will return the SSL v3 hash value */
    OPENSSL_cleanse(sm3tmp, sizeof(sm3tmp));

    return 1;

}

static const EVP_MD sm3_md = {
    NID_sm3,
    NID_SM3WithSM2, /* NID_sm3WithRSAEncryption replaced */
    SM3_DIGEST_LENGTH,
    0,
    init,
    update,
    final,
    NULL,
    NULL,
    SM3_CBLOCK,
    sizeof(EVP_MD *) + sizeof(SM3_CTX),
    ctrl
};

const EVP_MD *EVP_sm3(void)
{
    return &sm3_md;
}

#endif
